use itertools::Itertools;
use rand::prelude::*;
use rayon::prelude::*;

fn get_min_distance(nums: &[u32], target: u32, start: usize) -> Option<usize> {
    nums.iter()
        .enumerate()
        .filter(|&(_, e)| *e == target)
        .map(|(i, _)| (i as isize - start as isize).abs() as usize)
        .min()
}

fn get_min_distance2(nums: &[u32], target: u32, start: usize) -> Option<usize> {
    let n = nums.len();
    (start..n)
        .interleave((0..start).rev())
        .find(|&nums_index| nums[nums_index] == target)
        .map(|i| (i as isize - start as isize).abs() as usize)
}

fn par_get_min_distance(nums: &[u32], target: u32, start: usize) -> Option<usize> {
    nums.par_iter()
        .enumerate()
        .filter(|&(_, e)| *e == target)
        .map(|(i, _)| (i as isize - start as isize).abs() as usize)
        .min()
}

fn par_get_min_distance2(nums: &[u32], target: u32, start: usize) -> Option<usize> {
    let n = nums.len();
    (start..n)
        .into_par_iter()
        .interleave((0..start).into_par_iter().rev())
        .find_first(|&nums_index| nums[nums_index] == target)
        .map(|i| (i as isize - start as isize).abs() as usize)
}

fn random_input(limit: u32, length: usize) -> (Vec<u32>, u32, usize) {
    let nums: Vec<_> = std::iter::repeat_with(|| rand::random::<u32>() % limit)
        .take(length)
        .collect();
    let target = nums.choose(&mut rand::thread_rng()).copied().unwrap();
    let start = rand::random::<usize>() % length;
    (nums, target, start)
}

const BENCHES_NUM: usize = 100;

fn bench<F>(label: &str, op: F)
where
    F: Fn(&[u32], u32, usize) -> Option<usize>,
{
    let total_time: u128 = std::iter::repeat_with(|| {
        let (nums, target, start) = random_input(1_000_000, 100_000_000);
        let start_time = std::time::Instant::now();
        let r = op(&nums, target, start);
        let duration = start_time.elapsed();
        assert!(r.is_some());
        duration.as_micros()
    })
    .take(BENCHES_NUM)
    .sum();
    println!("{} took {}us", label, total_time / BENCHES_NUM as u128);
}

fn main() {
    let (nums, target, start) = random_input(10_000, 10_000_000);
    let r1 = get_min_distance(&nums, target, start);
    let r2 = get_min_distance2(&nums, target, start);
    assert_eq!(r1, r2);
    let r3 = par_get_min_distance(&nums, target, start);
    assert_eq!(r1, r3);
    let r4 = par_get_min_distance2(&nums, target, start);
    assert_eq!(r1, r4);

    bench("1", &get_min_distance);
    bench("2", &get_min_distance2);
    bench("par 1", &par_get_min_distance);
    bench("par 2", &par_get_min_distance2);
}
